/************************************************************
			Course		: 	CSC/ECE506
			Source 		:	mesi.cc
			Owner		:	Ed Gehringer
			Email Id	:	efg@ncsu.edu
*************************************************************/
		     
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
using namespace std;
#include "cache.h"

#include "main.h"
#include "mesi.h"
//
void MESI::PrRd(ulong addr, int processor_number) {
	// YOUR CODE HERE
	// The below comments are for reference and might not be
	// sufficient to match the debug runs. 
	// 
	// Update the Per-cache global counter to maintain LRU 
	// order among cache ways, updated on every cache access
	// Increment the global read counter.
	// Check whether the line is cached in the processor cache. 
	// If not cached, allocate a fresh line and update the state. (M,E,S,I) 
	// Check whether the directory entry is updated. If not updated,
	// create a fresh entry in the directory, update the sharer vector or list.
	// Update the directory state (U, EM, S_). 
	// Increment the directory operation counter like signalrds, 
	// response_replies etc... Invoke the relevant directory
	// signal functions like signalRd or signalRdX etc... 
	// If the line is cached in the processor cache, do not forget
	// to update the LRU
	// Do not forget to update miss/hit counter
	reads++;
	current_cycle++;
	cache_line * line = find_line(addr);

	if(!line)
	{
		read_misses++;
		cache_line * new_line = allocate_line(addr);
		if(!sharers(addr))
			{
			new_line->set_state(E);
			}
		
		else if(sharers(addr) > 0 )
			{
				new_line->set_state(S);

			}
		signalRd(addr, processor_number);

	
	}
	else{
			
		update_LRU(line);
	}
}
//
void MESI::PrWr(ulong addr, int processor_number) {
	// YOUR CODE HERE
	// Refer comments for PrRd
	current_cycle++;
	writes++;
	cache_line * line = find_line(addr);
	if(!line || line->get_state()==I)
	{
		write_misses++;
		cache_line * new_line = allocate_line(addr);
		new_line->set_state(M);
		signalRdX(addr, processor_number);
	}
	else
	{
		switch(line->get_state())
		{
			case M: update_LRU(line);
					break;
			case E: update_LRU(line);
					line->set_state(M);
					break;
			case S: line->set_state(M);
					update_LRU(line);
					signalUpgr(addr, processor_number);
					break;
			default: break;
		}
	}

}
//
cache_line * MESI::allocate_line(ulong addr) { 
    ulong tag;
    cache_state state;
   
    cache_line *victim = find_line_to_replace(addr);
    assert(victim != 0);
    state = victim->get_state();
    if (state == M){
        write_back(addr);

    }

	ulong victim_tag = victim->get_tag();	
	dir_entry* dir_line = directory->find_dir_line(victim_tag);
	if(dir_line!=NULL)	{
		dir_line->remove_sharer_entry(cache_num);
		int present = 0;
		present = dir_line->is_cached(num_processors);
		if(!present){
			dir_line->state = U;
		}
	}

    tag = tag_field(addr);   
    victim->set_tag(tag);
    victim->set_state(I);
    return victim;
}
//
void MESI::signalRd(ulong addr, int processor_number){
	// YOUR CODE HERE
	// The below comments are for reference and might not be
	// sufficient to match the debug runs.
	//
	// Check the directory state and update the cache2cache counter
	// Update the directory state
	// Update the vector/list
	// Send Intervention or Invalidation

	dir_entry *d_line = directory->find_dir_line(tag_field(addr));
	if(!d_line)
	{	
		d_line = directory->find_empty_line(tag_field(addr));

		d_line->set_dir_state(EM);
		d_line->add_sharer_entry(processor_number);
	}
	else
	{

		switch(d_line->get_state())
		{
			case U: 
					d_line->set_dir_state(EM);
					d_line->add_sharer_entry(processor_number);
					break; 
			case EM: cache2cache++; 
					 signal_rds++;
					 d_line->set_dir_state(S_);
					 d_line->add_sharer_entry(processor_number);
					 d_line->sendInt_to_sharer(addr, num_processors,processor_number);
					 break;
			case S_: signal_rds++;

					d_line->add_sharer_entry(processor_number);
					break;
			default: signal_rds++;
					break;
		}
	}

}

void MESI::signalRdX(ulong addr, int processor_number){
	// YOUR CODE HERE
	// Refer signalRd
	dir_entry * d_line = directory->find_dir_line(tag_field(addr));
	if(!d_line)
	{
		d_line = directory->find_empty_line(tag_field(addr));
		d_line->set_dir_state(EM);
		d_line->add_sharer_entry(processor_number);
	}
	else
	{
		switch(d_line->get_state())
		{
			case U: 
					d_line->set_dir_state(EM);
					d_line->add_sharer_entry(processor_number);
					break; 

			case EM: signal_rdxs++;
					 cache2cache++; 
					 d_line->add_sharer_entry(processor_number);
					 d_line->sendInv_to_sharer(addr, num_processors,processor_number);
					 break;
			case S_: signal_rdxs++;
					d_line->set_dir_state(EM);
					d_line->add_sharer_entry(processor_number);
					d_line->sendInv_to_sharer(addr, num_processors,processor_number);

			default: break;
		}
	}
}

void MESI::signalUpgr(ulong addr, int processor_number){
	// YOUR CODE HERE
	// Refer signalRd
	signal_upgrs++;
	dir_entry * d_line = directory->find_dir_line(tag_field(addr)); 
	if(
		d_line)
	{

		switch(d_line->get_state())
		{
			case S_: d_line->set_dir_state(EM);
				d_line->add_sharer_entry(processor_number);
				d_line-> sendInv_to_sharer(addr, num_processors, processor_number);
				break;
			default: break;
		}
		
	}
}

void MESI::Int(ulong addr) { 
	// YOUR CODE HERE
	// The below comments are for reference and might not be
	// sufficient to match the debug runs.
	//
	// Update the relevant counter, if the cache copy is dirty,
	// same needs to be written back to main memory. This is 
	// achieved by simply updating the writeback counter
	cache_line * line = find_line(addr);
	if(line)
	{

			switch(line->get_state())
		{
			case M: write_backs++;
					interventions++;
					break; 
			case E: write_backs++;
					interventions++;
					break;
			default: break;
		}
		line->set_state(S);
	}
}

void MESI::Inv(ulong addr) { 
	// YOUR CODE HERE
	// Refer Int
	cache_line * line = find_line(addr);
	if(line)
	{
		invalidations++;
		if(line->get_state()== M || line->get_state() == E)
		{
			write_backs++;
		}
		line->set_state(I);
	}
}
