#!/bin/bash

# Make results directory as needed
results_path="results/test"
ref_path="results/reference"

if [ ! -d "$results_path" ]; then
  echo "Creating directory structure: $results_path"
  mkdir -p "$results_path"
else
  echo "Directory structure already exists: $results_path"
fi

trace_path="traces/streamcluster.simdev.txt"

if [ -f "$trace_path" ]; then
  echo "Trace file found."
else
  echo "Trace file not found. Please add '$trace_path'."
  exit
fi

# Increasing Cache Size Tests
./dsm 131072 4 64 16 1 $trace_path > $results_path/ssci_cs_128.txt
./dsm 262144 4 64 16 1 $trace_path > $results_path/ssci_cs_256.txt
./dsm 524288 4 64 16 1 $trace_path > $results_path/ssci_cs_512.txt
./dsm 1048576 4 64 16 1 $trace_path > $results_path/ssci_cs_1024.txt
./dsm 2097152 4 64 16 1 $trace_path > $results_path/ssci_cs_2048.txt

echo "Differences for ssci_cs_128"
diff $results_path/ssci_cs_128.txt $ref_path/ssci_cs_128.txt
echo ""

echo "Differences for ssci_cs_256"
diff $results_path/ssci_cs_256.txt $ref_path/ssci_cs_256.txt
echo ""

echo "Differences for ssci_cs_512"
diff $results_path/ssci_cs_512.txt $ref_path/ssci_cs_512.txt
echo ""

echo "Differences for ssci_cs_1024"
diff $results_path/ssci_cs_1024.txt $ref_path/ssci_cs_1024.txt
echo ""

echo "Differences for ssci_cs_2048"
diff $results_path/ssci_cs_2048.txt $ref_path/ssci_cs_2048.txt
echo ""

# Increasing Associativity Tests
./dsm 1048576 1 64 16 1 $trace_path > $results_path/ssci_a_1.txt
./dsm 1048576 2 64 16 1 $trace_path > $results_path/ssci_a_2.txt
./dsm 1048576 4 64 16 1 $trace_path > $results_path/ssci_a_4.txt
./dsm 1048576 8 64 16 1 $trace_path > $results_path/ssci_a_8.txt

echo "Differences for ssci_a_1"
diff $results_path/ssci_a_1.txt $ref_path/ssci_a_1.txt
echo ""

echo "Differences for ssci_a_2"
diff $results_path/ssci_a_2.txt $ref_path/ssci_a_2.txt
echo ""

echo "Differences for ssci_a_4"
diff $results_path/ssci_a_4.txt $ref_path/ssci_a_4.txt
echo ""

echo "Differences for ssci_a_8"
diff $results_path/ssci_a_8.txt $ref_path/ssci_a_8.txt
echo ""

# Increasing Cache Block Size
./dsm 1048576 2 32 16 1 $trace_path > $results_path/ssci_cb_32.txt
./dsm 1048576 2 64 16 1 $trace_path > $results_path/ssci_cb_64.txt
./dsm 1048576 2 128 16 1 $trace_path > $results_path/ssci_cb_128.txt
./dsm 1048576 2 256 16 1 $trace_path > $results_path/ssci_cb_256.txt
./dsm 1048576 2 512 16 1 $trace_path > $results_path/ssci_cb_512.txt

echo "Differences for ssci_cb_32"
diff $results_path/ssci_cb_32.txt $ref_path/ssci_cb_32.txt
echo ""

echo "Differences for ssci_cb_64"
diff $results_path/ssci_cb_64.txt $ref_path/ssci_cb_64.txt
echo ""

echo "Differences for ssci_cb_128"
diff $results_path/ssci_cb_128.txt $ref_path/ssci_cb_128.txt
echo ""

echo "Differences for ssci_cb_256"
diff $results_path/ssci_cb_256.txt $ref_path/ssci_cb_256.txt
echo ""

echo "Differences for ssci_cb_512"
diff $results_path/ssci_cb_512.txt $ref_path/ssci_cb_512.txt
